#!/usr/bin/env python

print '\n'
print "##########################################################################"
print "This program implements algorithms 1 and 2 from \"A reconstruction problem"
print "for a class of phylogenetic networks with lateral gene transfers"
print "##########################################################################"
print '\n'
print '                         <<INSTRUCTIONS>>\n'
print """
1.  You must enter the eNewick strings of a set of trees to
    apply the Algorithm.
2.  First, enter the candidate to reduced principal subtree (T),
    and then each one of the candidates to reduced secondary
    subtree (T'_1, T'_2, ...).
3.  To stop entering secondary subtrees, give an empty string.
"""



from phylonetwork import PhyloNetwork,  LGTPhyloNetwork
import networkx as nx
import copy


# Obtain the input

T_=raw_input("Introduce the reduced principal subtree:")
T = LGTPhyloNetwork(eNewick=""+str(T_))
secondary_list=[]
T2_='start'
while T2_!= '':
	T2_ = raw_input("Introduce a new reduced secondary subtree:")
	if T2_!='':
		T2=LGTPhyloNetwork(eNewick=""+str(T2_))
		secondary_list.append(T2)




# Given a set s and a sorted chain (by inclusion), insert s (if is possible)
# in  the corresponding site.
def insert_set_chain(s,chain): 
    if not len(chain):
        return [s]+chain
    if set(s)<set(chain[0]):
        return [s]+chain
    if set(chain[len(chain)-1])<set(s):
        return chain+[s]
    for i in range(1,len(chain)):
        if (set(chain[i-1])<set(s))&(set(s)<set(chain[i])):
            chain.insert(i,s)
            return chain
    return 0
    
    
# Given C(T)\C(T') or C(T')\C(T) as input, return two segments U,W or U',W',
# respectively.
def obtain_segments(cluster_dif):
    #c1=[cluster_dif[0]]
    c1=[]
    c2=[]
    for i in range(0,len(cluster_dif)):
        s=insert_set_chain(cluster_dif[i],c1)
        if s==0:
            t=insert_set_chain(cluster_dif[i],c2)
            if t==0:
                return 'Error: Condition (a) fails. It is not possible to obtain exactly two segments!'
            else:
                c2=t
        else:
            c1=s
    return c1,c2


# Check if the intersection of minimums and if the length of the 
# segments in condition (a) of Proposition 8 is satisfied.
def check_a_old(seg_P, seg_S):
    cont=False
    k=len(seg_P[0])
    l=len(seg_P[1])
    k_t=len(seg_S[0])
    l_t=len(seg_S[1])
    if (k+l==0) or (k_t+l_t==0):
    	return False
    if l!=0:
       if len(list(set(seg_P[0][k-1]) & set(seg_P[1][l-1])))!=0:
            return False
    if l_t!=0:
        if len(list(set(seg_S[0][k_t-1]) & set(seg_S[1][l_t-1])))!=0:
            return False    
    if ((k_t== k+1)&(l_t==l-1)) or ((k_t== l+1)&(l_t==k-1)) or ((k_t== k-1)&(l_t==l+1)) or ((k_t== l-1)&(l_t==k+1)) or ((k_t== k)&(l_t==l+1)) or ((l_t== k)&(k_t==l+1)) or ((k_t== l)&(l_t==k+1)) or ((l_t== l)&(k_t==k+1)) or ((l_t== l)&(k_t==k-1)) or ((k_t== l)&(l_t==k-1)) or ((l_t== k)&(k_t==l-1)) or ((k_t== k)&(l_t==l-1)) or ((k_t== k)&(l_t==l)) or ((l_t== k)&(k_t==l)):
        return True
    return cont


# Check if the intersection of minimums and if the length of the
# segments in condition (a) of Proposition 8 is satisfied.
def check_a(seg_P, seg_S):
    cont=False
    k=len(seg_P[0])
    l=len(seg_P[1])
    k_t=len(seg_S[0])
    l_t=len(seg_S[1])
    if (k+l==0) or (k_t+l_t==0):
    	return False
    if l!=0:
        if len(list(set(seg_P[0][k-1]) & set(seg_P[1][l-1])))!=0:
            return False
    if l_t!=0:
        if len(list(set(seg_S[0][k_t-1]) & set(seg_S[1][l_t-1])))!=0:
            return False
    if ((k>1)&(l==0))&((k_t>1)&(l_t==0)):
        return False
    if ((k_t== k+1)&(l_t==l-1)) or ((k_t== l+1)&(l_t==k-1)) or ((k_t== k-1)&(l_t==l+1)) or ((k_t== l-1)&(l_t==k+1)) or ((k_t== k)&(l_t==l+1)) or ((l_t== k)&(k_t==l+1)) or ((k_t== l)&(l_t==k+1)) or ((l_t== l)&(k_t==k+1)) or ((l_t== l)&(k_t==k-1)) or ((k_t== l)&(l_t==k-1)) or ((l_t== k)&(k_t==l-1)) or ((k_t== k)&(l_t==l-1)) or ((k_t== k)&(l_t==l)) or ((l_t== k)&(k_t==l)):
        return True
    return cont
    
    
# Check if condition (b) of Proposition 8 is satisfied.
# If it is, return the cluster B.
def check_b(p_c,s_c, min_p, min_s):
    aux_p=[]
    aux_s=[]
    for i in p_c:
        aux_p.append(set(i))
    for i in s_c:
        aux_s.append(set(i))    
    cont=True
    B=tuple(set(min_p) & set(min_s)) 
    if ((set(B) in aux_p)==False) or ((set(B) in aux_s)==False):
        cont=False
    return cont,B 
    

# Check if condition (c) of Proposition 8 is satisfied.
# If it is, return the clusters A_k.
def check_c(p_c,s_c, min_p, B, uk_hybrid):
    aux_p=[]
    aux_s=[]
    for i in p_c:
        aux_p.append(set(i))
    for i in s_c:
        aux_s.append(set(i)) 
    cont=True
    Ak=tuple(set(min_p) - set(B))
    #Cl=tuple(set(min_s) - set(B)) 
    if ((set(Ak) in aux_s)==False):
        cont=False
    if (uk_hybrid==False):
    	if ((set(Ak) in aux_p)==False):
    		cont=False
    if (uk_hybrid==True):
    	if ((set(Ak) in aux_p)==True):
    		cont=False
    
    return cont,Ak 
   
   
# Check if condition (d) of Proposition 8 is satisfied.
# If it is, return the clusters C_l.
def check_d(p_c,s_c, min_s, B, w_hybrid):
    aux_p=[]
    aux_s=[]
    for i in p_c:
        aux_p.append(set(i))
    for i in s_c:
        aux_s.append(set(i)) 
    cont=True
    #Ak=tuple(set(min_p) - set(B))
    Cl=tuple(set(min_s) - set(B)) 
    if ((set(Cl) in aux_p)==False):
        cont=False
    if (w_hybrid==False):
    	if ((set(Cl) in aux_s)==False):
    		cont=False
    if (w_hybrid==True):
    	if ((set(Cl) in aux_s)==True):
    		cont=False
    
    return cont,Cl 
    
    
# Check if condition (e) of Proposition 8 is satisfied.
def check_e(Ak,seg_U,seg_U_t):
    cont=True
    k=len(seg_U)
    k_t=len(seg_U_t)
    if k<2:
        return True
    if (k==k_t+1) or (k==k_t):
        if (k==k_t+1):
            if not set(Ak)<set(seg_U_t[0]):
                return False
            seg_U_t=[Ak]+seg_U_t
        else:
            if not set(Ak)<set(seg_U_t[1]):
                return False
        for i in range(1,k): 
            if (set(seg_U[k-i])-set(seg_U[k-(i+1)])) != (set(seg_U_t[k-i])-set(seg_U_t[k-(i+1)])):
                return False
    else:
        return False
    return cont 
    
# Check if condition (f) of Proposition 8 is satisfied.
def check_f(Cl,seg_W,seg_W_t):
    cont=True
    l=len(seg_W)
    l_t=len(seg_W_t)
    if l<2:
        return True
    if (l_t==l+1) or (l_t==l):
        if (l_t==l+1):
            if not set(Cl)<set(seg_W[0]):
                return False
            seg_W=[Cl]+seg_W
        else:
            if not set(Cl)<set(seg_W[1]):
                return False
        for i in range(1,l_t): 
            if (set(seg_W[l_t-i])-set(seg_W[l_t-(i+1)])) != (set(seg_W_t[l_t-i])-set(seg_W_t[l_t-(i+1)])):
                return False
    else:
        return False
    return cont 
    
    
    
    
# Detect segments U, W, U' and W'
def distinguish_segments(ch_in_p,ch_in_s):
    k=len(ch_in_p[0])
    l=len(ch_in_p[1])
    l_=len(ch_in_s[0])
    k_=len(ch_in_s[1])
    k0_is_k=False
    l0_is_l=False

    if ((k==1)&(l==0)): 
        U=ch_in_p[0]
        W=ch_in_p[1] 
        if set(ch_in_s[0][0])<set(U[0]):
            U_=ch_in_s[0]
            W_=ch_in_s[1]
        else:
            W_=ch_in_s[0]
            U_=ch_in_s[1]
        if len(U)==len(U_):
            k0_is_k=True
    if ((k>1)&(l==0)):
        U=ch_in_p[0]
        W=ch_in_p[1]
        if set(ch_in_s[0][0])<set(U[1]):
            U_=ch_in_s[0]
            W_=ch_in_s[1]
        else:
            U_=ch_in_s[1]
            W_=ch_in_s[0]
        if len(U_)==len(U):
            k0_is_k=True
    if ((l_==1)&(k_==0)&(k==1)&(l==1)):
        W_=ch_in_s[0]
        U_=ch_in_s[1]
        l0_is_l=True
        if set(ch_in_p[0][0])<set(W_[0]):
            W=ch_in_p[0]
            U=ch_in_p[1]
        else:
            W=ch_in_p[1]
            U=ch_in_p[0]
    if ((l_>1)&(k_==0)):
        W_=ch_in_s[0]
        U_=ch_in_s[1]
        if set(ch_in_p[0][0])<set(W_[1]):
            W=ch_in_p[0]
            U=ch_in_p[1]
        else:
            W=ch_in_p[1]
            U=ch_in_p[0]
        if len(W_)==len(W):
            l0_is_l=True 
    if (k>0)&(l>0)&(k_>0)&(l_>0):
        if (len(set(ch_in_p[0][0])&set(ch_in_s[0][0]))!=0) & (len(set(ch_in_p[0][0])&set(ch_in_s[1][0]))!=0):
            U=ch_in_p[0]
            W=ch_in_p[1]            
        else:
            U=ch_in_p[1]
            W=ch_in_p[0]
        if (len(set(ch_in_s[0][0])&set(U[0]))!=0) & (len(set(ch_in_s[0][0])&set(W[0]))!=0):
            W_=ch_in_s[0]
            U_=ch_in_s[1]
        else:
            W_=ch_in_s[1]
            U_=ch_in_s[0]
        if len(U)==len(U_):
            k0_is_k=True
        if len(W)==len(W_):
            l0_is_l=True
            
    return U,W,U_,W_,k0_is_k,l0_is_l
           



# If condition (a) is satisfied, check all the others conditions from (b) to (f).
def prop10(princ_cluster, sec_cluster, seg_princ, seg_sec):
    S=distinguish_segments(seg_princ,seg_sec)
    U=S[0]
    W=S[1]
    U_t=S[2]
    W_t=S[3]
    uk_hyb=S[4]
    w_hyb=S[5]
    min_in_T=U[0]
    min_in_T2=W_t[0]
    ch_b=check_b(princ_cluster, sec_cluster, min_in_T, min_in_T2)
    B=ch_b[1]
    if ch_b[0]:
        ch_c=check_c(princ_cluster, sec_cluster, min_in_T, B, uk_hyb)
        A_k=ch_c[1]
        if ch_c[0]:
            ch_d=check_d(princ_cluster, sec_cluster, min_in_T2, B, w_hyb)
            C_l=ch_d[1]
            if ch_d[0]:
                ch_e=check_e(A_k,U,U_t)
                if ch_e:
                    ch_f=check_f(C_l,W,W_t)
                    if ch_f:
                        return B,C_l,w_hyb
                    else:
                        return 'Error: Condition (f) fails.'
                else:
                    return 'Error: Condition (e) fails.'
            else:
                return 'Error: Condition (d) fails.'
        else:
            return 'Error: Condition (c) fails.'
    else:
        return 'Error: Condition (b) fails.'
    

# Return the node with an specific cluster c in the tree T.
def node_with_cluster(c,T):
    for i in T.nodes():
        if set(T.cluster(i))==set(c):
            return i
    return 'This cluster is not the cluster of any node in the tree.'
    

# Relabel the node which becomes hybrid
def change_to_hybrid(net,u,sec_par,i):
    N=copy.deepcopy(net)
    new='#'+i
    N.add_node(new)
    for e in net.edges(data=True):
        if e[0]==u:
            if e[2]['secondary']==True:
                N.add_edge(new,e[1],attr_dict={'secondary':True})
            else:
                N.add_edge(new,e[1],attr_dict={'secondary':False})
            N.remove_edge(e[0],e[1])
        if e[1]==u:
            if e[2]['secondary']==True:
                N.add_edge(e[0],new,attr_dict={'secondary':True})
            else:
                N.add_edge(e[0],new,attr_dict={'secondary':False})
            N.remove_edge(e[0],e[1])
    N.add_edge(sec_par,new,attr_dict={'secondary':True})      
        
    if net.is_leaf(u):
        lab=net.label(u)
        N.set_label(new,lab)
    N._labels[u]=None
    N.remove_node(u)
    N.cache={}
    return N
    
    
# Returns the principal cluster of a node in a net  
def cluster_p(u,N):
    T=N.principal_subtree(simplify=False)
    return T.cluster(u)
    

# Returns the node with a given principal cluster in a net
def node_with_princ_cluster(c,N):
    for i in N.nodes():
        if set(cluster_p(i,N))==set(c):
            return i
    return 'This cluster is not the cluster of a node in the tree.'


# Returns True if a node is principally elementary
def principally_elementary(u,N):
    T=N.principal_subtree(simplify=False)
    return T.is_elementary_node(u)



# Given the cluster B and C_l, applies an iteration in the Algorithm 2. 
def add_edge(Net,u_star,h,i,w_hybrid):
    N=copy.deepcopy(Net) 
    if w_hybrid:
        #N.add_edge(u_star,h,attr_dict={'secondary':True})
        new_node=u_star
    else:
        for e in N.edges(data=True):
            if (e[1]==u_star)& (e[2]['secondary']==False):
                p_u_star=e[0] 
        #if N.is_elementary_node(p_u_star):
        if principally_elementary(p_u_star,N)==True:
            #N.add_edge(p_u_star,h,attr_dict={'secondary':True})
            new_node=p_u_star
        else: 
            new_node=N._generate_new_id()
            N.add_node(new_node)
            N.add_edge(p_u_star,new_node,attr_dict={'secondary':False})
            N.add_edge(new_node,u_star, attr_dict={'secondary':False})
            N.remove_edge(p_u_star,u_star) 
            #N.add_edge(new_node,h,attr_dict={'secondary':True})
    N=change_to_hybrid(N,h,new_node,i)
    N.cache = {}
    return N


# Algorithm 1
def Algorithm1(T,T2):
    princ_clust=T.cluster_representation()
    sec_clust=T2.cluster_representation()
    seg_princ=obtain_segments(list(set(princ_clust)-set(sec_clust)))
    seg_sec=obtain_segments(list(set(sec_clust)-set(princ_clust)))
    if check_a(seg_princ,seg_sec):
        res=prop10(princ_clust, sec_clust, seg_princ, seg_sec)
        u_star=res[1]
        h=res[0]
        return add_edge(T,node_with_cluster(u_star,T),node_with_cluster(h,T))
    else:
        return 'Error'


# Algorithm 2
def Algorithm2(T,sec_subtrees):
    N=copy.deepcopy(T)
    princ_clust=T.cluster_representation()
    i=1
    for T2 in sec_subtrees:
        sec_clust=T2.cluster_representation()
        seg_princ=obtain_segments(list(set(princ_clust)-set(sec_clust)))
        seg_sec=obtain_segments(list(set(sec_clust)-set(princ_clust)))
        if check_a(seg_princ,seg_sec):
            res=prop10(princ_clust, sec_clust, seg_princ, seg_sec)
            u_star=res[1]
            h=res[0]
            N=add_edge(N,node_with_princ_cluster(u_star,N),node_with_princ_cluster(h,N),str(i),res[2])
        else:
            return 'Error: Condition (a) fails. The four segments of clusters do not satisfy the lenght conditions, or the intersection of minumums is not empty!'
    	i+=1
    if len(list(nx.simple_cycles(N)))==0:
        return N
    else:
        return 'Error: The output graph contains a cycle which is '+ str(list(nx.simple_cycles(N))[0])
 
# used to obtain the extended Newick of a node
def eNewickLGT_node(net,u,visited):
    if net.is_leaf(u):
        #return self._labels[u]
        mylabel = net._labels.get(u,'')
        if net.is_hybrid_node(u):
            mylabel+=u
        return mylabel
    
    if u in visited:
        return u
    visited.append(u)
    
    children_str = []
    for v in net.successors(u):
        if (u,v) in net.principal_edges():
            child_str = eNewickLGT_node(net,v,visited)
        else:
            child_str = v[:1] + 'LGT' + v[1:]
        children_str.append(child_str)
        

    children_str.sort()
    internal=','.join(children_str)
    mylabel=net.label(u) or ''
    if net.is_hybrid_node(u):
        #print "hibrid"
        mylabel+=u
    return '('+internal+')'+mylabel
    
# Return the extended Newick string of a LGT network    
def eNewickLGT(net):
    visited = []
    string = ''
    for root in net.roots():
        string += eNewickLGT_node(net,root,visited)+';'
    return string


# The output
 
print '\n'
print '                         <<OUTPUT>>\n'

R=Algorithm2(T,secondary_list)
if ';' in str(R):
	print 'The eNewick string of the restricted LGT network is: '       
	print eNewickLGT(R)
else:
	print R


